/* 
 * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#ifndef CRMD__H
#  define CRMD__H

#  define SYS_NAME     CRM_SYSTEM_CRMD
#  define PID_FILE     WORKING_DIR "/"SYS_NAME".pid"
#  define DAEMON_LOG   DEVEL_DIR"/"SYS_NAME".log"
#  define DAEMON_DEBUG DEVEL_DIR"/"SYS_NAME".debug"

extern GMainLoop *crmd_mainloop;
extern GHashTable *ipc_clients;

extern void crmd_metadata(void);

#endif
